@@ -1,24 +1,7 @@
 /*
-    Copyright (C) 2016 Volker Krause <vkrause@kde.org>
+    SPDX-FileCopyrightText: 2016 Volker Krause <vkrause@kde.org>
 
-    Permission is hereby granted, free of charge, to any person obtaining
-    a copy of this software and associated documentation files (the
-    "Software"), to deal in the Software without restriction, including
-    without limitation the rights to use, copy, modify, merge, publish,
-    distribute, sublicense, and/or sell copies of the Software, and to
-    permit persons to whom the Software is furnished to do so, subject to
-    the following conditions:
-
-    The above copyright notice and this permission notice shall be included
-    in all copies or substantial portions of the Software.
-
-    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
-    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
-    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
-    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
-    CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
-    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
-    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
+    SPDX-License-Identifier: MIT
 */
 
 #include "test-config.h"
@@ -33,8 +16,10 @@
 #include <QDir>
 #include <QFile>
 #include <QObject>
+#include <QTest>
 #include <QTextStream>
-#include <qtest.h>
+
+#include <map>
 
 using namespace KSyntaxHighlighting;
 
@@ -49,7 +34,9 @@ public:
             return;
         }
         m_out.setDevice(&outFile);
+#if QT_VERSION < QT_VERSION_CHECK(6, 0, 0)
         m_out.setCodec("UTF-8");
+#endif
 
         QFile f(inFileName);
         if (!f.open(QFile::ReadOnly)) {
@@ -58,7 +45,9 @@ public:
         }
 
         QTextStream in(&f);
+#if QT_VERSION < QT_VERSION_CHECK(6, 0, 0)
         in.setCodec("UTF-8");
+#endif
         State state;
         while (!in.atEnd()) {
             m_currentLine = in.readLine();
@@ -70,12 +59,13 @@ public:
     }
 
 protected:
-    void applyFormat(int offset, int length, const Format &format) Q_DECL_OVERRIDE
+    void applyFormat(int offset, int length, const Format &format) override
     {
-        if (format.name().isEmpty())
-            m_out << "<dsNormal>" << m_currentLine.midRef(offset, length) << "</dsNormal>";
-        else
-            m_out << "<" << format.name() << ">" << m_currentLine.midRef(offset, length) << "</" << format.name() << ">";
+        if (format.name().isEmpty()) {
+            m_out << "<dsNormal>" << QStringView(m_currentLine).mid(offset, length) << "</dsNormal>";
+        } else {
+            m_out << "<" << format.name() << ">" << QStringView(m_currentLine).mid(offset, length) << "</" << format.name() << ">";
+        }
     }
 
 private:
@@ -83,20 +73,24 @@ private:
     QString m_currentLine;
 };
 
-
 class TestHighlighterTest : public QObject
 {
     Q_OBJECT
 public:
-    explicit TestHighlighterTest(QObject *parent = nullptr) : QObject(parent), m_repo(nullptr) {}
+    explicit TestHighlighterTest(QObject *parent = nullptr)
+        : QObject(parent)
+        , m_repo(nullptr)
+    {
+    }
+
 private:
-        Repository *m_repo;
-        QSet<QString> m_coveredDefinitions;
+    Repository *m_repo;
+    std::map<QString, QStringList> m_coveredDefinitions;
 
 private Q_SLOTS:
     void initTestCase()
     {
-        QStandardPaths::enableTestMode(true);
+        QStandardPaths::setTestModeEnabled(true);
         m_repo = new Repository;
         initRepositorySearchPaths(*m_repo);
     }
@@ -109,18 +103,40 @@ private Q_SLOTS:
         QVERIFY(uncoveredList.open(QFile::WriteOnly));
 
         int count = 0;
-        foreach (const auto &def, m_repo->definitions()) {
-            if (def.isHidden() || !def.isValid())
+        for (const auto &def : m_repo->definitions()) {
+            if (!def.isValid()) {
                 continue;
+            }
             ++count;
-            if (m_coveredDefinitions.contains(def.name()))
+            if (m_coveredDefinitions.find(def.name()) != m_coveredDefinitions.end()) {
                 coveredList.write(def.name().toUtf8() + '\n');
-            else
+            } else {
                 uncoveredList.write(def.name().toUtf8() + '\n');
+            }
         }
 
         qDebug() << "Syntax definitions with test coverage:" << ((float)m_coveredDefinitions.size() * 100.0f / (float)count) << "%";
 
+        // we don't want multiple tests for the same highlighting
+        // tests should be consolidated into one useful file per highlighting
+        // the update script for https://kate-editor.org/syntax/ will check that no duplicated output is there, too
+        bool duplicates = false;
+        for (const auto &entry : std::as_const(m_coveredDefinitions)) {
+            if (entry.second.size() <= 1) {
+                continue;
+            }
+
+            // abort and tell about duplicated test cases!
+            qWarning() << "Multiple unit tests for the language " << entry.first;
+            for (const auto &testCase : entry.second) {
+                qWarning() << "  - " << testCase;
+            }
+            duplicates = true;
+        }
+        if (duplicates) {
+            QFAIL("Multiple unit tests for the same language found, see for details the output above!");
+        }
+
         delete m_repo;
         m_repo = nullptr;
     }
@@ -133,23 +149,30 @@ private Q_SLOTS:
         QTest::addColumn<QString>("syntax");
 
         const QDir dir(QStringLiteral(TESTSRCDIR "/input"));
-        foreach (const auto &fileName, dir.entryList(QDir::Files | QDir::NoSymLinks | QDir::Readable, QDir::Name)) {
-            const auto inFile = dir.absoluteFilePath(fileName);
-            if (inFile.endsWith(QLatin1String(".syntax")))
+        for (const auto &fileName : dir.entryList(QDir::Files | QDir::NoSymLinks | QDir::Readable | QDir::Hidden, QDir::Name)) {
+            // skip .clang-format file we use to avoid formatting test files
+            if (fileName == QLatin1String(".clang-format")) {
                 continue;
+            }
+
+            const auto inFile = dir.absoluteFilePath(fileName);
+            if (inFile.endsWith(QLatin1String(".syntax"))) {
+                continue;
+            }
 
             QString syntax;
             QFile syntaxOverride(inFile + QStringLiteral(".syntax"));
-            if (syntaxOverride.exists() && syntaxOverride.open(QFile::ReadOnly))
+            if (syntaxOverride.exists() && syntaxOverride.open(QFile::ReadOnly)) {
                 syntax = QString::fromUtf8(syntaxOverride.readAll()).trimmed();
+            }
 
-            QTest::newRow(fileName.toUtf8().constData()) << inFile
-                << (QStringLiteral(TESTBUILDDIR "/output/") + fileName + QStringLiteral(".ref"))
-                << (QStringLiteral(TESTSRCDIR "/reference/") + fileName + QStringLiteral(".ref"))
-                << syntax;
+            QTest::newRow(fileName.toUtf8().constData()) << inFile << (QStringLiteral(TESTBUILDDIR "/output/") + fileName + QStringLiteral(".ref"))
+                                                         << (QStringLiteral(TESTSRCDIR "/reference/") + fileName + QStringLiteral(".ref")) << syntax;
         }
 
-        QVERIFY(QDir().mkpath(QStringLiteral(TESTBUILDDIR "/output/")));
+        // cleanup before we test
+        QDir(QStringLiteral(TESTBUILDDIR "/output/")).removeRecursively();
+        QDir().mkpath(QStringLiteral(TESTBUILDDIR "/output/"));
     }
 
     void testHighlight()
@@ -161,8 +184,9 @@ private Q_SLOTS:
         QVERIFY(m_repo);
 
         auto def = m_repo->definitionForFileName(inFile);
-        if (!syntax.isEmpty())
+        if (!syntax.isEmpty()) {
             def = m_repo->definitionForName(syntax);
+        }
 
         TestHighlighter highlighter;
         highlighter.setTheme(m_repo->defaultTheme());
@@ -170,7 +194,7 @@ private Q_SLOTS:
 
         QVERIFY(def.isValid());
         qDebug() << "Using syntax" << def.name();
-        m_coveredDefinitions.insert(def.name());
+        m_coveredDefinitions[def.name()].push_back(inFile);
         highlighter.setDefinition(def);
         highlighter.highlightFile(inFile, outFile);
 
@@ -179,11 +203,8 @@ private Q_SLOTS:
          */
         compareFiles(refFile, outFile);
     }
-
 };
 
 QTEST_GUILESS_MAIN(TestHighlighterTest)
 
 #include "testhighlighter.moc"
-
-
