@@ -14,6 +14,7 @@
 #include "context_p.h"
 #include "format.h"
 #include "format_p.h"
+#include "highlightingdata_p.hpp"
 #include "ksyntaxhighlighting_logging.h"
 #include "ksyntaxhighlighting_version.h"
 #include "repository.h"
@@ -39,10 +40,7 @@ DefinitionData::DefinitionData()
 {
 }
 
-DefinitionData::~DefinitionData()
-{
-    qDeleteAll(contexts);
-}
+DefinitionData::~DefinitionData() = default;
 
 DefinitionData *DefinitionData::get(const Definition &def)
 {
@@ -51,30 +49,21 @@ DefinitionData *DefinitionData::get(const Definition &def)
 
 Definition::Definition()
     : d(new DefinitionData)
-{
-}
-
-Definition::Definition(const Definition &other)
-    : d(other.d)
 {
     d->q = *this;
 }
 
+Definition::Definition(Definition &&other) noexcept = default;
+Definition::Definition(const Definition &) = default;
+Definition::~Definition() = default;
+Definition &Definition::operator=(Definition &&other) noexcept = default;
+Definition &Definition::operator=(const Definition &) = default;
+
 Definition::Definition(std::shared_ptr<DefinitionData> &&dd)
     : d(std::move(dd))
 {
 }
 
-Definition::~Definition()
-{
-}
-
-Definition &Definition::operator=(const Definition &rhs)
-{
-    d = rhs.d;
-    return *this;
-}
-
 bool Definition::operator==(const Definition &other) const
 {
     return d->fileName == other.d->fileName;
@@ -223,8 +212,9 @@ bool Definition::setKeywordList(const QString &name, const QStringList &content)
     if (list) {
         list->setKeywordList(content);
         return true;
-    } else
+    } else {
         return false;
+    }
 }
 
 QVector<Format> Definition::formats() const
@@ -245,45 +235,23 @@ QVector<Definition> Definition::includedDefinitions() const
     d->load();
 
     // init worklist and result used as guard with this definition
-    QVector<Definition> queue{*this};
+    QVector<const DefinitionData *> queue{d.get()};
     QVector<Definition> definitions{*this};
-    while (!queue.isEmpty()) {
-        // Iterate all context rules to find associated Definitions. This will
-        // automatically catch other Definitions referenced with IncludeRuldes or ContextSwitch.
-        const auto definition = queue.takeLast();
-        for (const auto &context : qAsConst(definition.d->contexts)) {
-            // handle context switch attributes of this context itself
-            for (const auto switchContext :
-                 {context->lineEndContext().context(), context->lineEmptyContext().context(), context->fallthroughContext().context()}) {
-                if (switchContext) {
-                    if (!definitions.contains(switchContext->definition())) {
-                        queue.push_back(switchContext->definition());
-                        definitions.push_back(switchContext->definition());
-                    }
-                }
-            }
-
-            // handle the embedded rules
-            for (const auto &rule : context->rules()) {
-                // handle include rules like inclusion
-                if (!definitions.contains(rule->definition())) {
-                    queue.push_back(rule->definition());
-                    definitions.push_back(rule->definition());
-                }
-
-                // handle context switch context inclusion
-                if (auto switchContext = rule->context().context()) {
-                    if (!definitions.contains(switchContext->definition())) {
-                        queue.push_back(switchContext->definition());
-                        definitions.push_back(switchContext->definition());
-                    }
-                }
+    while (!queue.empty()) {
+        const auto *def = queue.back();
+        queue.pop_back();
+        for (const auto &defRef : def->immediateIncludedDefinitions) {
+            const auto definition = defRef.definition();
+            if (!definitions.contains(definition)) {
+                definitions.push_back(definition);
+                queue.push_back(definition.d.get());
             }
         }
     }
 
     // remove the 1st entry, since it is this Definition
-    definitions.pop_front();
+    definitions.front() = std::move(definitions.back());
+    definitions.pop_back();
 
     return definitions;
 }
@@ -312,17 +280,18 @@ QVector<QPair<QChar, QString>> Definition::characterEncodings() const
     return d->characterEncodings;
 }
 
-Context *DefinitionData::initialContext() const
+Context *DefinitionData::initialContext()
 {
-    Q_ASSERT(!contexts.isEmpty());
-    return contexts.first();
+    Q_ASSERT(!contexts.empty());
+    return &contexts.front();
 }
 
-Context *DefinitionData::contextByName(const QString &wantedName) const
+Context *DefinitionData::contextByName(const QString &wantedName)
 {
-    for (const auto context : contexts) {
-        if (context->name() == wantedName)
-            return context;
+    for (auto &context : contexts) {
+        if (context.name() == wantedName) {
+            return &context;
+        }
     }
     return nullptr;
 }
@@ -336,37 +305,43 @@ KeywordList *DefinitionData::keywordList(const QString &wantedName)
 Format DefinitionData::formatByName(const QString &wantedName) const
 {
     const auto it = formats.constFind(wantedName);
-    if (it != formats.constEnd())
+    if (it != formats.constEnd()) {
         return it.value();
+    }
 
     return Format();
 }
 
 bool DefinitionData::isLoaded() const
 {
-    return !contexts.isEmpty();
+    return !contexts.empty();
 }
 
 bool DefinitionData::load(OnlyKeywords onlyKeywords)
 {
-    if (fileName.isEmpty())
+    if (fileName.isEmpty()) {
         return false;
+    }
 
-    if (isLoaded())
+    if (isLoaded()) {
         return true;
+    }
 
-    if (bool(onlyKeywords) && keywordIsLoaded)
+    if (bool(onlyKeywords) && keywordIsLoaded) {
         return true;
+    }
 
     QFile file(fileName);
-    if (!file.open(QFile::ReadOnly))
+    if (!file.open(QFile::ReadOnly)) {
         return false;
+    }
 
     QXmlStreamReader reader(&file);
     while (!reader.atEnd()) {
         const auto token = reader.readNext();
-        if (token != QXmlStreamReader::StartElement)
+        if (token != QXmlStreamReader::StartElement) {
             continue;
+        }
 
         if (reader.name() == QLatin1String("highlighting")) {
             loadHighlighting(reader, onlyKeywords);
@@ -375,25 +350,16 @@ bool DefinitionData::load(OnlyKeywords onlyKeywords)
             }
         }
 
-        else if (reader.name() == QLatin1String("general"))
+        else if (reader.name() == QLatin1String("general")) {
             loadGeneral(reader);
+        }
     }
 
     for (auto it = keywordLists.begin(); it != keywordLists.end(); ++it) {
         it->setCaseSensitivity(caseSensitive);
     }
 
-    for (const auto context : qAsConst(contexts)) {
-        context->resolveContexts();
-        context->resolveIncludes();
-        context->resolveAttributeFormat();
-    }
-
-    for (const auto context : qAsConst(contexts)) {
-        for (const auto &rule : context->rules()) {
-            rule->resolvePostProcessing();
-        }
-    }
+    resolveContexts();
 
     return true;
 }
@@ -402,9 +368,21 @@ void DefinitionData::clear()
 {
     // keep only name and repo, so we can re-lookup to make references persist over repo reloads
     keywordLists.clear();
-    qDeleteAll(contexts);
     contexts.clear();
     formats.clear();
+    contextDatas.clear();
+    immediateIncludedDefinitions.clear();
+    wordDelimiters = WordDelimiters();
+    wordWrapDelimiters = wordDelimiters;
+    keywordIsLoaded = false;
+    hasFoldingRegions = false;
+    indentationBasedFolding = false;
+    foldingIgnoreList.clear();
+    singleLineCommentMarker.clear();
+    singleLineCommentPosition = CommentPosition::StartOfLine;
+    multiLineCommentStartMarker.clear();
+    multiLineCommentEndMarker.clear();
+    characterEncodings.clear();
 
     fileName.clear();
     section.clear();
@@ -414,8 +392,6 @@ void DefinitionData::clear()
     license.clear();
     mimetypes.clear();
     extensions.clear();
-    wordDelimiters = WordDelimiters();
-    wordWrapDelimiters = wordDelimiters;
     caseSensitive = Qt::CaseSensitive;
     version = 0.0f;
     priority = 0;
@@ -427,14 +403,16 @@ bool DefinitionData::loadMetaData(const QString &definitionFileName)
     fileName = definitionFileName;
 
     QFile file(definitionFileName);
-    if (!file.open(QFile::ReadOnly))
+    if (!file.open(QFile::ReadOnly)) {
         return false;
+    }
 
     QXmlStreamReader reader(&file);
     while (!reader.atEnd()) {
         const auto token = reader.readNext();
-        if (token != QXmlStreamReader::StartElement)
+        if (token != QXmlStreamReader::StartElement) {
             continue;
+        }
         if (reader.name() == QLatin1String("language")) {
             return loadLanguage(reader);
         }
@@ -457,11 +435,13 @@ bool DefinitionData::loadMetaData(const QString &file, const QCborMap &obj)
     fileName = file;
 
     const auto exts = obj.value(QLatin1String("extensions")).toString();
-    for (const auto &ext : exts.split(QLatin1Char(';'), Qt::SkipEmptyParts))
+    for (const auto &ext : exts.split(QLatin1Char(';'), Qt::SkipEmptyParts)) {
         extensions.push_back(ext);
+    }
     const auto mts = obj.value(QLatin1String("mimetype")).toString();
-    for (const auto &mt : mts.split(QLatin1Char(';'), Qt::SkipEmptyParts))
+    for (const auto &mt : mts.split(QLatin1Char(';'), Qt::SkipEmptyParts)) {
         mimetypes.push_back(mt);
+    }
 
     return true;
 }
@@ -471,8 +451,9 @@ bool DefinitionData::loadLanguage(QXmlStreamReader &reader)
     Q_ASSERT(reader.name() == QLatin1String("language"));
     Q_ASSERT(reader.tokenType() == QXmlStreamReader::StartElement);
 
-    if (!checkKateVersion(reader.attributes().value(QLatin1String("kateversion"))))
+    if (!checkKateVersion(reader.attributes().value(QLatin1String("kateversion")))) {
         return false;
+    }
 
     name = reader.attributes().value(QLatin1String("name")).toString();
     section = reader.attributes().value(QLatin1String("section")).toString();
@@ -485,13 +466,16 @@ bool DefinitionData::loadLanguage(QXmlStreamReader &reader)
     author = reader.attributes().value(QLatin1String("author")).toString();
     license = reader.attributes().value(QLatin1String("license")).toString();
     const auto exts = reader.attributes().value(QLatin1String("extensions")).toString();
-    for (const auto &ext : exts.split(QLatin1Char(';'), Qt::SkipEmptyParts))
+    for (const auto &ext : exts.split(QLatin1Char(';'), Qt::SkipEmptyParts)) {
         extensions.push_back(ext);
+    }
     const auto mts = reader.attributes().value(QLatin1String("mimetype")).toString();
-    for (const auto &mt : mts.split(QLatin1Char(';'), Qt::SkipEmptyParts))
+    for (const auto &mt : mts.split(QLatin1Char(';'), Qt::SkipEmptyParts)) {
         mimetypes.push_back(mt);
-    if (reader.attributes().hasAttribute(QLatin1String("casesensitive")))
+    }
+    if (reader.attributes().hasAttribute(QLatin1String("casesensitive"))) {
         caseSensitive = Xml::attrToBool(reader.attributes().value(QLatin1String("casesensitive"))) ? Qt::CaseSensitive : Qt::CaseInsensitive;
+    }
     return true;
 }
 
@@ -555,14 +539,14 @@ void DefinitionData::loadContexts(QXmlStreamReader &reader)
     Q_ASSERT(reader.name() == QLatin1String("contexts"));
     Q_ASSERT(reader.tokenType() == QXmlStreamReader::StartElement);
 
+    contextDatas.reserve(32);
+
     while (!reader.atEnd()) {
         switch (reader.tokenType()) {
         case QXmlStreamReader::StartElement:
             if (reader.name() == QLatin1String("context")) {
-                auto context = new Context;
-                context->setDefinition(q);
-                context->load(reader);
-                contexts.push_back(context);
+                contextDatas.push_back(HighlightingContextData());
+                contextDatas.back().load(name, reader);
             }
             reader.readNext();
             break;
@@ -575,6 +559,50 @@ void DefinitionData::loadContexts(QXmlStreamReader &reader)
     }
 }
 
+void DefinitionData::resolveContexts()
+{
+    contexts.reserve(contextDatas.size());
+
+    /**
+     * Transform all HighlightingContextData to Context.
+     * This is necessary so that Context::resolveContexts() can find the referenced contexts.
+     */
+    for (const auto &contextData : std::as_const(contextDatas)) {
+        contexts.emplace_back(*this, contextData);
+    }
+
+    /**
+     * Resolves contexts and rules.
+     */
+    auto ctxIt = contexts.begin();
+    for (const auto &contextData : std::as_const(contextDatas)) {
+        ctxIt->resolveContexts(*this, contextData);
+        ++ctxIt;
+    }
+
+    /**
+     * To free the memory, constDatas is emptied because it is no longer used.
+     */
+    contextDatas.clear();
+    contextDatas.shrink_to_fit();
+
+    /**
+     * Resolved includeRules.
+     */
+    for (auto &context : contexts) {
+        context.resolveIncludes(*this);
+    }
+
+    /**
+     * Post-processing on rules.
+     */
+    for (const auto &context : contexts) {
+        for (auto &rule : context.rules()) {
+            rule->resolvePostProcessing();
+        }
+    }
+}
+
 void DefinitionData::loadItemData(QXmlStreamReader &reader)
 {
     Q_ASSERT(reader.name() == QLatin1String("itemDatas"));
@@ -618,24 +646,25 @@ void DefinitionData::loadGeneral(QXmlStreamReader &reader)
             ++elementRefCounter;
 
             if (reader.name() == QLatin1String("keywords")) {
-                if (reader.attributes().hasAttribute(QLatin1String("casesensitive")))
+                if (reader.attributes().hasAttribute(QLatin1String("casesensitive"))) {
                     caseSensitive = Xml::attrToBool(reader.attributes().value(QLatin1String("casesensitive"))) ? Qt::CaseSensitive : Qt::CaseInsensitive;
+                }
 
                 // adapt wordDelimiters
                 wordDelimiters.append(reader.attributes().value(QLatin1String("additionalDeliminator")));
                 wordDelimiters.remove(reader.attributes().value(QLatin1String("weakDeliminator")));
 
                 // adapt WordWrapDelimiters
-                auto wordWrapDeliminatorAttr = reader.attributes().value(
-                    QLatin1String("wordWrapDeliminator"));
-                if (wordWrapDeliminatorAttr.isEmpty())
+                auto wordWrapDeliminatorAttr = reader.attributes().value(QLatin1String("wordWrapDeliminator"));
+                if (wordWrapDeliminatorAttr.isEmpty()) {
                     wordWrapDelimiters = wordDelimiters;
-                else {
+                } else {
                     wordWrapDelimiters.append(wordWrapDeliminatorAttr);
                 }
             } else if (reader.name() == QLatin1String("folding")) {
-                if (reader.attributes().hasAttribute(QLatin1String("indentationsensitive")))
+                if (reader.attributes().hasAttribute(QLatin1String("indentationsensitive"))) {
                     indentationBasedFolding = Xml::attrToBool(reader.attributes().value(QLatin1String("indentationsensitive")));
+                }
             } else if (reader.name() == QLatin1String("emptyLines")) {
                 loadFoldingIgnoreList(reader);
             } else if (reader.name() == QLatin1String("comments")) {
@@ -649,8 +678,9 @@ void DefinitionData::loadGeneral(QXmlStreamReader &reader)
             break;
         case QXmlStreamReader::EndElement:
             --elementRefCounter;
-            if (elementRefCounter == 0)
+            if (elementRefCounter == 0) {
                 return;
+            }
             reader.readNext();
             break;
         default:
@@ -688,8 +718,9 @@ void DefinitionData::loadComments(QXmlStreamReader &reader)
             break;
         case QXmlStreamReader::EndElement:
             --elementRefCounter;
-            if (elementRefCounter == 0)
+            if (elementRefCounter == 0) {
                 return;
+            }
             reader.readNext();
             break;
         default:
@@ -719,8 +750,9 @@ void DefinitionData::loadFoldingIgnoreList(QXmlStreamReader &reader)
             break;
         case QXmlStreamReader::EndElement:
             --elementRefCounter;
-            if (elementRefCounter == 0)
+            if (elementRefCounter == 0) {
                 return;
+            }
             reader.readNext();
             break;
         default:
@@ -754,8 +786,9 @@ void DefinitionData::loadSpellchecking(QXmlStreamReader &reader)
             break;
         case QXmlStreamReader::EndElement:
             --elementRefCounter;
-            if (elementRefCounter == 0)
+            if (elementRefCounter == 0) {
                 return;
+            }
             reader.readNext();
             break;
         default:
@@ -789,16 +822,25 @@ quint16 DefinitionData::foldingRegionId(const QString &foldName)
     return RepositoryPrivate::get(repo)->foldingRegionId(name, foldName);
 }
 
-DefinitionRef::DefinitionRef()
+void DefinitionData::addImmediateIncludedDefinition(const Definition &def)
 {
+    if (get(def) != this) {
+        DefinitionRef defRef(def);
+        if (!immediateIncludedDefinitions.contains(defRef)) {
+            immediateIncludedDefinitions.push_back(std::move(defRef));
+        }
+    }
 }
 
+DefinitionRef::DefinitionRef() = default;
+
 DefinitionRef::DefinitionRef(const Definition &def)
     : d(def.d)
 {
 }
 
-DefinitionRef::~DefinitionRef()
+DefinitionRef::DefinitionRef(Definition &&def)
+    : d(std::move(def.d))
 {
 }
 
@@ -808,10 +850,17 @@ DefinitionRef &DefinitionRef::operator=(const Definition &def)
     return *this;
 }
 
+DefinitionRef &DefinitionRef::operator=(Definition &&def)
+{
+    d = std::move(def.d);
+    return *this;
+}
+
 Definition DefinitionRef::definition() const
 {
-    if (!d.expired())
+    if (!d.expired()) {
         return Definition(d.lock());
+    }
     return Definition();
 }
 
